// SPDX-FileCopyrightText: 2023 KylinSoft Co., Ltd.
//
// SPDX-License-Identifier: MulanPSL-2.0
#include <assert.h>
#include <stdlib.h>

#include <wlr/types/wlr_compositor.h>

#include <kywc/log.h>

#include "input/seat.h"
#include "output.h"
#include "server.h"
#include "view/workspace.h"
#include "view_p.h"

static struct view_manager *view_manager = NULL;

struct view *view_manager_get_activated(void)
{
    return view_manager->activated.view;
}

struct view_layer *view_manager_get_layer(enum layer layer, bool in_workspace)
{
    if (!in_workspace) {
        return &view_manager->layers[layer];
    }

    switch (layer) {
    case LAYER_BELOW:
    case LAYER_NORMAL:
    case LAYER_ABOVE:
        return workspace_layer(workspace_manager_get_current(), layer);
    default:
        return &view_manager->layers[layer];
    }
}

static void view_update_output(struct view *view)
{
    int lx = 0, ly = 0;

    if (view->base.mapped) {
        struct kywc_box *geo = &view->base.geometry;
        /* udpate view most-at output */
        lx = geo->x + geo->width / 2;
        ly = geo->y + geo->height / 2;
    } else if (view->output) {
        /* no need to update output when unmapped except output was destroyed */
        return;
    }

    struct kywc_output *old = view->output;
    if (!old || old->destroying || !old->state.enabled ||
        !kywc_box_contains_point(&output_from_kywc_output(old)->geometry, lx, ly)) {
        view->output = kywc_output_at_point(lx, ly);
    }

    if (old != view->output) {
        wl_list_remove(&view->output_destroy.link);
        if (view->output) {
            wl_signal_add(&view->output->events.destroy, &view->output_destroy);
        } else {
            wl_list_init(&view->output_destroy.link);
        }
        wl_signal_emit_mutable(&view->events.output, old);
    }
}

static void view_handle_output_destroy(struct wl_listener *listener, void *data)
{
    struct view *view = wl_container_of(listener, view, output_destroy);
    view->output = NULL;
    view_update_output(view);
}

void view_move_to_output(struct view *view, struct kywc_output *kywc_output)
{
    if (view->output == kywc_output) {
        return;
    }

    struct kywc_view *kywc_view = &view->base;
    struct output *dst = output_from_kywc_output(kywc_output);
    struct output *src = output_from_kywc_output(view->output);
    struct kywc_box *dst_box = &dst->usable_area;
    struct kywc_box *src_box = &src->usable_area;
    double frac_x = (double)dst_box->width / src_box->width;
    double frac_y = (double)dst_box->height / src_box->height;

    if (kywc_view->fullscreen) {
        kywc_view_resize(kywc_view, &dst->geometry);
        return;
    }

    if (kywc_view->maximized) {
        struct kywc_box geo;
        view_get_tiled_geometry(view, &geo, kywc_output, KYWC_TILE_ALL);
        kywc_view_resize(kywc_view, &geo);
        return;
    }

    if (kywc_view->tiled) {
        struct kywc_box geo;
        view_get_tiled_geometry(view, &geo, kywc_output, kywc_view->tiled);
        kywc_view_resize(kywc_view, &geo);
        return;
    }

    /* actual view geomtry with margin */
    int x = kywc_view->geometry.x - kywc_view->margin.off_x;
    int y = kywc_view->geometry.y - kywc_view->margin.off_y;
    int w = kywc_view->geometry.width + kywc_view->margin.off_width;
    int h = kywc_view->geometry.height + kywc_view->margin.off_height;

    int nx, ny;
    /* keep align to edges */
    if (src_box->x == x) {
        nx = dst_box->x + kywc_view->margin.off_x;
    } else if (src_box->x + src_box->width == x + w) {
        nx = dst_box->x + dst_box->width - w + kywc_view->margin.off_x;
    } else {
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
        nx = ceil(MAX(x - src_box->x, 0) * frac_x) + dst_box->x + kywc_view->margin.off_x;
    }
    if (src_box->y == y) {
        ny = dst_box->y + kywc_view->margin.off_y;
    } else if (src_box->y + src_box->height == y + h) {
        ny = dst_box->y + dst_box->height - h + kywc_view->margin.off_y;
    } else {
        ny = ceil(MAX(y - src_box->y, 0) * frac_y) + dst_box->y + kywc_view->margin.off_y;
    }
#undef MAX
    /* move to dst */
    kywc_view_move(kywc_view, nx, ny);
}

void view_init(struct view *view, const struct view_impl *impl, void *data)
{
    struct kywc_view *kywc_view = &view->base;
    wlr_addon_set_init(&view->addons);

    wl_signal_init(&kywc_view->events.premap);
    wl_signal_init(&kywc_view->events.map);
    wl_signal_init(&kywc_view->events.unmap);
    wl_signal_init(&kywc_view->events.destroy);
    wl_signal_init(&kywc_view->events.activate);
    wl_signal_init(&kywc_view->events.maximize);
    wl_signal_init(&kywc_view->events.minimize);
    wl_signal_init(&kywc_view->events.fullscreen);
    wl_signal_init(&kywc_view->events.tile);
    wl_signal_init(&kywc_view->events.title);
    wl_signal_init(&kywc_view->events.app_id);
    wl_signal_init(&kywc_view->events.position);
    wl_signal_init(&kywc_view->events.size);
    wl_signal_init(&kywc_view->events.decoration);
    wl_signal_init(&kywc_view->events.shadow);

    view->impl = impl;
    view->data = data;
    wl_list_init(&view->children);
    wl_list_init(&view->parent_link);
    wl_list_init(&view->view_proxies);
    wl_signal_init(&view->events.parent);
    wl_signal_init(&view->events.workspace);
    wl_signal_init(&view->events.output);

    kywc_view->minimizable = true;
    kywc_view->maximizable = true;
    kywc_view->fullscreenable = true;
    kywc_view->closeable = true;
    kywc_view->movable = true;
    kywc_view->resizable = true;
    kywc_view->activatable = true;
    kywc_view->focusable = true;
    kywc_view->shadeable = true;

    /* create view tree and disable it */
    struct view_layer *layer = view_manager_get_layer(LAYER_NORMAL, true);
    view->tree = ky_scene_tree_create(layer->tree);
    ky_scene_node_set_enabled(ky_scene_node_from_tree(view->tree), false);
    view->content = ky_scene_tree_create(view->tree);

    struct output *output = input_current_output(input_manager_get_default_seat());
    view->output = &output->base;
    view->output_destroy.notify = view_handle_output_destroy;
    wl_signal_add(&view->output->events.destroy, &view->output_destroy);

    view_set_workspace(view, workspace_manager_get_current());

    wl_signal_emit_mutable(&view_manager->events.new_view, kywc_view);
}

void view_get_tiled_geometry(struct view *view, struct kywc_box *geometry,
                             struct kywc_output *kywc_output, enum kywc_tile tile)
{
    struct output *output = output_from_kywc_output(kywc_output);
    struct kywc_box *usable = &output->usable_area;
    struct kywc_view *kywc_view = &view->base;
    int32_t min_width = kywc_view->min_width;
    int32_t min_height = kywc_view->min_height;
    int32_t half_width = usable->width / 2;
    int32_t half_height = usable->height / 2;
    bool use_min_width = half_width < min_width;
    bool use_min_height = half_height < min_height;

    switch (tile) {
    case KYWC_TILE_NONE:
        *geometry = view->saved.geometry;
        return;
    case KYWC_TILE_TOP:
        geometry->x = usable->x + kywc_view->margin.off_x;
        geometry->y = usable->y + kywc_view->margin.off_y;
        geometry->width = usable->width - kywc_view->margin.off_width;
        geometry->height = use_min_height ? min_height : half_height - kywc_view->margin.off_height;
        return;
    case KYWC_TILE_BOTTOM:
        geometry->x = usable->x + kywc_view->margin.off_x;
        geometry->y = use_min_height ? usable->y + usable->height - min_height -
                                           (kywc_view->margin.off_height - kywc_view->margin.off_y)
                                     : usable->y + half_height + kywc_view->margin.off_y;
        geometry->width = usable->width - kywc_view->margin.off_width;
        geometry->height = use_min_height ? min_height : half_height - kywc_view->margin.off_height;
        return;
    case KYWC_TILE_LEFT:
        geometry->x = usable->x + kywc_view->margin.off_x;
        geometry->y = usable->y + kywc_view->margin.off_y;
        geometry->width = use_min_width ? min_width : half_width - kywc_view->margin.off_width;
        geometry->height = usable->height - kywc_view->margin.off_height;
        return;
    case KYWC_TILE_RIGHT:
        geometry->x = use_min_width ? usable->x + usable->width - min_width -
                                          (kywc_view->margin.off_width - kywc_view->margin.off_x)
                                    : usable->x + half_width + kywc_view->margin.off_x;
        geometry->y = usable->y + kywc_view->margin.off_y;
        geometry->width = use_min_width ? min_width : half_width - kywc_view->margin.off_width;
        geometry->height = usable->height - kywc_view->margin.off_height;
        return;
    case KYWC_TILE_TOP_LEFT:
        geometry->x = usable->x + kywc_view->margin.off_x;
        geometry->y = usable->y + kywc_view->margin.off_y;
        geometry->width = use_min_width ? min_width : half_width - kywc_view->margin.off_width;
        geometry->height = use_min_height ? min_height : half_height - kywc_view->margin.off_height;
        return;
    case KYWC_TILE_BOTTOM_LEFT:
        geometry->x = usable->x + kywc_view->margin.off_x;
        geometry->y = use_min_height ? usable->y + usable->height - min_height -
                                           (kywc_view->margin.off_height - kywc_view->margin.off_y)
                                     : usable->y + half_height + kywc_view->margin.off_y;
        geometry->width = use_min_width ? min_width : half_width - kywc_view->margin.off_width;
        geometry->height = use_min_height ? min_height : half_height - kywc_view->margin.off_height;
        return;
    case KYWC_TILE_TOP_RIGHT:
        geometry->x = use_min_width ? usable->x + usable->width - min_width -
                                          (kywc_view->margin.off_width - kywc_view->margin.off_x)
                                    : usable->x + half_width + kywc_view->margin.off_x;
        geometry->y = usable->y + kywc_view->margin.off_y;
        geometry->width = use_min_width ? min_width : half_width - kywc_view->margin.off_width;
        geometry->height = use_min_height ? min_height : half_height - kywc_view->margin.off_height;
        return;
    case KYWC_TILE_BOTTOM_RIGHT:
        geometry->x = use_min_width ? usable->x + usable->width - min_width -
                                          (kywc_view->margin.off_width - kywc_view->margin.off_x)
                                    : usable->x + half_width + kywc_view->margin.off_x;
        geometry->y = use_min_height ? usable->y + usable->height - min_height -
                                           (kywc_view->margin.off_height - kywc_view->margin.off_y)
                                     : usable->y + half_height + kywc_view->margin.off_y;
        geometry->width = use_min_width ? min_width : half_width - kywc_view->margin.off_width;
        geometry->height = use_min_height ? min_height : half_height - kywc_view->margin.off_height;
        return;
    case KYWC_TILE_CENTER:
        geometry->x = use_min_width ? usable->x + half_width - min_width / 2 -
                                          (kywc_view->margin.off_width - kywc_view->margin.off_x)
                                    : usable->x + usable->width / 4 + kywc_view->margin.off_x;
        geometry->y = use_min_height ? usable->y + half_height - min_height / 2 -
                                           (kywc_view->margin.off_height - kywc_view->margin.off_y)
                                     : usable->y + usable->height / 4 + kywc_view->margin.off_y;
        geometry->width = use_min_width ? min_width : half_width - kywc_view->margin.off_width;
        geometry->height = use_min_height ? min_height : half_height - kywc_view->margin.off_height;
        return;
    case KYWC_TILE_ALL:
        geometry->x = usable->x + kywc_view->margin.off_x;
        geometry->y = usable->y + kywc_view->margin.off_y;
        geometry->width = usable->width - kywc_view->margin.off_width;
        geometry->height = usable->height - kywc_view->margin.off_height;
        return;
    }
}

void view_map(struct view *view)
{
    struct kywc_view *kywc_view = &view->base;

    wl_signal_emit_mutable(&kywc_view->events.premap, NULL);

    /* assume that request_minimize may emited before map */
    struct ky_scene_node *node = ky_scene_node_from_tree(view->tree);
    ky_scene_node_set_enabled(node, !kywc_view->minimized);

    kywc_view->mapped = true;

    if (view->pending.action) {
        /* add a fallback geometry */
        if (kywc_view->maximized || kywc_view->fullscreen) {
            view_get_tiled_geometry(view, &view->saved.geometry, view->output, KYWC_TILE_CENTER);
        }
        if (view->impl->configure) {
            view->impl->configure(view);
        }
    }

    kywc_view->has_initial_position = false;
    kywc_view_activate(kywc_view);
    seat_focus_surface(input_manager_get_default_seat(), view->surface);

    kywc_log(KYWC_DEBUG, "kywc_view %p map", kywc_view);
    wl_signal_emit_mutable(&kywc_view->events.map, NULL);

    if (view->current_proxy) {
        struct view_proxy *proxy;
        wl_list_for_each(proxy, &view->view_proxies, view_link) {
            wl_signal_emit_mutable(&proxy->workspace->events.view_enter, view);
        }
        if (!kywc_view->minimized && view_manager->show_desktop_enabled) {
            view_manager_show_desktop(false, false);
        }
    }
}

void view_unmap(struct view *view)
{
    struct kywc_view *kywc_view = &view->base;

    kywc_view->title = kywc_view->app_id = NULL;
    ky_scene_node_set_enabled(ky_scene_node_from_tree(view->tree), false);
    kywc_view->mapped = false;

    if (view->pending.configure_timeout) {
        wl_event_source_remove(view->pending.configure_timeout);
        view->pending.configure_timeout = NULL;
    }

    wl_list_remove(&view->parent_link);
    wl_list_init(&view->parent_link);
    view->parent = NULL;

    struct view *child, *tmp;
    wl_list_for_each_safe(child, tmp, &view->children, parent_link) {
        wl_list_remove(&child->parent_link);
        wl_list_init(&child->parent_link);
        child->parent = NULL;
    }

    kywc_log(KYWC_DEBUG, "kywc_view %p unmap", kywc_view);
    input_rebase_all_cursor();
    wl_signal_emit_mutable(&kywc_view->events.unmap, NULL);

    if (view->current_proxy) {
        struct view_proxy *proxy;
        wl_list_for_each(proxy, &view->view_proxies, view_link) {
            wl_signal_emit_mutable(&proxy->workspace->events.view_leave, view);
        }
    }
}

#define CONFIGURE_TIMEOUT_MS 100

static int view_handle_configure_timeout(void *data)
{
    struct view *view = data;

    kywc_log(KYWC_INFO, "client (%s) did not respond to configure request %d in %d ms",
             view->base.app_id, view->pending.configure_serial, CONFIGURE_TIMEOUT_MS);

    /* fallback for pending actions */
    if (view_action_change_size(view->pending.configure_action)) {
        struct kywc_box *current = &view->base.geometry;
        struct kywc_box *pending = &view->pending.configure_geometry;
        int x = pending->x, y = pending->y;
        /* fix wobbling when resize */
        if (view->pending.configure_action & VIEW_ACTION_RESIZE) {
            if (current->x != pending->x) {
                x += pending->width - current->width;
            }
            if (current->y != pending->y) {
                y += pending->height - current->height;
            }
        }
        view_helper_move(view, x, y);
    }

    view_configured(view);

    return 0;
}

void view_configure(struct view *view, uint32_t serial)
{
    view->pending.configure_serial = serial;
    view->pending.configure_action |= view->pending.action;
    view->pending.configure_geometry = view->pending.geometry;

    view->pending.action = VIEW_ACTION_NOP;
    view->pending.geometry = (struct kywc_box){ 0 };

    if (serial == 0) {
        return;
    }

    if (!view->pending.configure_timeout) {
        view->pending.configure_timeout = wl_event_loop_add_timer(
            view_manager->server->event_loop, view_handle_configure_timeout, view);
    }

    wl_event_source_timer_update(view->pending.configure_timeout, CONFIGURE_TIMEOUT_MS);
}

void view_configured(struct view *view)
{
    struct kywc_view *kywc_view = &view->base;

    if (view->pending.configure_action != VIEW_ACTION_NOP &&
        (view->pending.configure_action & VIEW_ACTION_RESIZE) == 0) {
        input_rebase_all_cursor();
    }

    if (view->pending.configure_action & VIEW_ACTION_FULLSCREEN) {
        wl_signal_emit_mutable(&kywc_view->events.fullscreen, NULL);
    }

    if (view->pending.configure_action & VIEW_ACTION_MAXIMIZE) {
        wl_signal_emit_mutable(&kywc_view->events.maximize, NULL);
    }

    if (view->pending.configure_action & VIEW_ACTION_TILE) {
        wl_signal_emit_mutable(&kywc_view->events.tile, NULL);
    }

    if (view->pending.configure_timeout) {
        wl_event_source_remove(view->pending.configure_timeout);
        view->pending.configure_timeout = NULL;
    }

    view->pending.configure_serial = 0;
    view->pending.configure_action = VIEW_ACTION_NOP;
    view->pending.configure_geometry = (struct kywc_box){ 0 };
}

void view_proxy_destroy(struct view_proxy *view_proxy)
{
    if (!view_proxy) {
        return;
    }
    if (view_proxy->view->base.mapped) {
        wl_signal_emit_mutable(&view_proxy->workspace->events.view_leave, view_proxy->view);
    }
    wl_list_remove(&view_proxy->view_link);
    wl_list_remove(&view_proxy->workspace_link);
    ky_scene_node_destroy(ky_scene_node_from_tree(view_proxy->tree));
    free(view_proxy);
}

void view_set_current_proxy(struct view *view, struct view_proxy *view_proxy)
{
    if (view->current_proxy == view_proxy) {
        return;
    }
    if (!view_proxy) {
        assert(wl_list_empty(&view->view_proxies));
    }
    if (view_proxy) {
        ky_scene_node_reparent(ky_scene_node_from_tree(view->tree), view_proxy->tree);
    }
    view->current_proxy = view_proxy;
    wl_signal_emit_mutable(&view->events.workspace, NULL);
}

static void view_proxies_destroy(struct view *view)
{
    struct view_proxy *proxy, *tmp;
    wl_list_for_each_safe(proxy, tmp, &view->view_proxies, view_link) {
        view_proxy_destroy(proxy);
    }
    view_set_current_proxy(view, NULL);
}

void view_destroy(struct view *view)
{
    struct kywc_view *kywc_view = &view->base;
    kywc_log(KYWC_DEBUG, "kywc_view %p destroy", kywc_view);

    wlr_addon_set_finish(&view->addons);
    wl_signal_emit_mutable(&kywc_view->events.destroy, NULL);

    wl_list_remove(&view->output_destroy.link);

    ky_scene_node_destroy(ky_scene_node_from_tree(view->tree));
    view_proxies_destroy(view);

    view->impl->destroy(view);
}

void view_set_title(struct view *view, const char *title)
{
    struct kywc_view *kywc_view = &view->base;

    kywc_view->title = title;
    kywc_log(KYWC_DEBUG, "kywc_view %p title: %s", kywc_view, title);

    wl_signal_emit_mutable(&kywc_view->events.title, NULL);
}

void view_set_app_id(struct view *view, const char *app_id)
{
    struct kywc_view *kywc_view = &view->base;

    kywc_view->app_id = app_id;
    kywc_log(KYWC_DEBUG, "kywc_view %p app_id: %s", kywc_view, app_id);

    wl_signal_emit_mutable(&kywc_view->events.app_id, NULL);
}

void view_set_decoration(struct view *view, enum kywc_ssd ssd)
{
    struct kywc_view *kywc_view = &view->base;
    if (kywc_view->ssd == ssd) {
        return;
    }

    kywc_view->ssd = ssd;
    kywc_log(KYWC_DEBUG, "kywc_view %p need ssd %d", kywc_view, ssd);

    wl_signal_emit_mutable(&kywc_view->events.decoration, NULL);
}

void view_set_shaded(struct view *view, bool shaded)
{
    struct kywc_view *kywc_view = &view->base;
    shaded = kywc_view->shadeable && shaded;
    if (kywc_view->shaded == shaded) {
        return;
    }

    kywc_view->shaded = shaded;
    kywc_log(KYWC_DEBUG, "kywc_view %p need shadow %d", kywc_view, shaded);

    wl_signal_emit_mutable(&kywc_view->events.shadow, NULL);
}

struct view_proxy *view_proxy_by_workspace(struct view *view, struct workspace *workspace)
{
    struct view_proxy *view_proxy;
    wl_list_for_each(view_proxy, &view->view_proxies, view_link) {
        if (view_proxy->workspace == workspace) {
            return view_proxy;
        }
    }
    return NULL;
}

static struct view_proxy *view_proxy_create(struct view *view, struct workspace *workspace)
{
    struct view_proxy *proxy = calloc(1, sizeof(struct view_proxy));
    if (!proxy) {
        return NULL;
    }
    proxy->view = view;
    proxy->workspace = workspace;

    wl_list_insert(&workspace->view_proxies, &proxy->workspace_link);
    wl_list_insert(&view->view_proxies, &proxy->view_link);

    /* create view_proxy tree from new workspace view_layer */
    enum layer layer =
        view->base.kept_above ? LAYER_ABOVE : (view->base.kept_below ? LAYER_BELOW : LAYER_NORMAL);
    struct view_layer *view_layer = workspace_layer(workspace, layer);
    proxy->tree = ky_scene_tree_create(view_layer->tree);
    if (view->base.mapped) {
        wl_signal_emit_mutable(&workspace->events.view_enter, view);
    }
    return proxy;
}

static void view_do_set_workspace(struct view *view, struct workspace *workspace)
{
    if (!workspace && !view->current_proxy) {
        return;
    }
    if (!workspace) {
        /* set workspace to null must reparent view tree first */
        view_proxies_destroy(view);
        wl_list_init(&view->view_proxies);
        return;
    }
    struct view_proxy *proxy = view_proxy_by_workspace(view, workspace);
    if (!proxy) {
        proxy = view_proxy_create(view, workspace);
        if (!proxy) {
            return;
        }
    }

    if (proxy->tree != ky_scene_node_get_parent(ky_scene_node_from_tree(view->tree))) {
        view_set_current_proxy(view, proxy);
    }

    // destroy all proxies except the new worskpace
    struct view_proxy *view_proxy, *tmp;
    wl_list_for_each_safe(view_proxy, tmp, &view->view_proxies, view_link) {
        if (view_proxy != proxy) {
            view_proxy_destroy(view_proxy);
        }
    }
    view->show_in_all_workspaces = false;
}

void view_set_workspace(struct view *view, struct workspace *workspace)
{
    assert(workspace);

    view_do_set_workspace(view, workspace);
}

void view_unset_workspace(struct view *view, struct view_layer *layer)
{
    assert(layer->layer != LAYER_BELOW && layer->layer != LAYER_NORMAL &&
           layer->layer != LAYER_ABOVE);

    ky_scene_node_reparent(ky_scene_node_from_tree(view->tree), layer->tree);
    view_do_set_workspace(view, NULL);
}

struct view_proxy *view_add_workspace(struct view *view, struct workspace *workspace)
{
    if (!view || !workspace) {
        return NULL;
    }
    struct view_proxy *view_proxy = view_proxy_by_workspace(view, workspace);
    if (view_proxy) {
        return NULL;
    }
    return view_proxy_create(view, workspace);
}

void view_remove_workspace(struct view *view, struct workspace *workspace)
{
    if (!view || !workspace) {
        return;
    }
    struct view_proxy *view_proxy = view_proxy_by_workspace(view, workspace);
    if (!view_proxy) {
        return;
    }

    struct view_proxy *proxy;
    wl_list_for_each_reverse(proxy, &view->view_proxies, view_link) {
        if (proxy != view_proxy) {
            break;
        }
    }
    /* add to all workspace if the view only exists in this workspace */
    if (&proxy->view_link == &view->view_proxies) {
        view_add_all_workspace(view);
        return;
    }
    /* del the proxy if view exists in multi workspaces */
    if (view->current_proxy == view_proxy) {
        view_set_current_proxy(view, proxy);
    }
    view_proxy_destroy(view_proxy);
    view->show_in_all_workspaces = false;
}

void view_add_all_workspace(struct view *view)
{
    if (!view) {
        return;
    }
    int workspace_num = workspace_manager_get_count();
    struct workspace *workspace = NULL;
    for (int i = 0; i < workspace_num; i++) {
        workspace = workspace_by_position(i);
        view_add_workspace(view, workspace);
    }
    view->show_in_all_workspaces = true;
}

void view_set_parent(struct view *view, struct view *parent)
{
    if (view->parent == parent) {
        return;
    }

    wl_list_remove(&view->parent_link);
    if (parent) {
        wl_list_insert(&parent->children, &view->parent_link);
    } else {
        wl_list_init(&view->parent_link);
    }

    kywc_log(KYWC_DEBUG, "view %p set parent to %p", view, parent);

    view->parent = parent;
    wl_signal_emit_mutable(&view->events.parent, NULL);
}

void kywc_view_add_new_listener(struct wl_listener *listener)
{
    wl_signal_add(&view_manager->events.new_view, listener);
}

struct view *view_from_kywc_view(struct kywc_view *kywc_view)
{
    struct view *view = wl_container_of(kywc_view, view, base);
    return view;
}

struct view *view_try_from_wlr_surface(struct wlr_surface *wlr_surface)
{
    return wlr_surface->data;
}

void kywc_view_close(struct kywc_view *kywc_view)
{
    struct view *view = view_from_kywc_view(kywc_view);

    /* it is not allowed to close a view that has children views */
    if (!wl_list_empty(&view->children)) {
        kywc_log(KYWC_WARN, "close a view that still have children views");
        return;
    }

    if (view->impl->close) {
        view->impl->close(view);
    }
}

void kywc_view_move(struct kywc_view *kywc_view, int x, int y)
{
    struct view *view = view_from_kywc_view(kywc_view);

    view->pending.action |= VIEW_ACTION_MOVE;
    view->pending.geometry.x = x;
    view->pending.geometry.y = y;

    if (kywc_view->mapped && view->impl->configure) {
        view->impl->configure(view);
    }
}

void kywc_view_resize(struct kywc_view *kywc_view, struct kywc_box *geometry)
{
    struct view *view = view_from_kywc_view(kywc_view);

    view->pending.action |= VIEW_ACTION_RESIZE;
    view->pending.geometry = *geometry;

    if (kywc_view->mapped && view->impl->configure) {
        view->impl->configure(view);
    }
}

static void view_set_activated(struct view *view, bool activated);

static void handle_activated_view_minimized(struct wl_listener *listener, void *data)
{
    struct view *view = view_manager->activated.view;
    if (!view->base.minimized) {
        return;
    }

    view_set_activated(view, false);
    view_topmost_activate(workspace_manager_get_current());
}

static void handle_activated_view_unmap(struct wl_listener *listener, void *data)
{
    struct view *view = view_manager->activated.view;
    view_set_activated(view, false);
    view_topmost_activate(workspace_manager_get_current());
}

static void view_reparent_fullscreen(struct view *view, bool active);

static void view_set_activated(struct view *view, bool activated)
{
    struct kywc_view *kywc_view = &view->base;
    if (kywc_view->activated == activated) {
        return;
    }

    if (activated) {
        /* listen activated view's minimize and unmap signals,
         * so that we can auto activate another view.
         */
        view_manager->activated.view = view;
        wl_signal_add(&kywc_view->events.minimize, &view_manager->activated.minimize);
        wl_signal_add(&kywc_view->events.destroy, &view_manager->activated.unmap);
    } else {
        wl_list_remove(&view_manager->activated.minimize.link);
        wl_list_remove(&view_manager->activated.unmap.link);
        view_manager->activated.view = NULL;
    }

    /* change fullscreen layer when activation changed */
    if (kywc_view->fullscreen) {
        view_reparent_fullscreen(view, activated);
    }

    if (kywc_view->minimized && activated) {
        kywc_view_set_minimized(kywc_view, false);
    }

    kywc_view->activated = activated;
    view->pending.action |= VIEW_ACTION_ACTIVATE;

    if (kywc_view->mapped && view->impl->configure) {
        view->impl->configure(view);
    }

    wl_signal_emit_mutable(&kywc_view->events.activate, NULL);
}

static void view_activate(struct view *view)
{
    struct view *last = view_manager->activated.view;
    if (!view->base.activatable || last == view) {
        return;
    }

    if (last) {
        view_set_activated(last, false);
    }

    view_set_activated(view, true);
}

void view_topmost_activate(struct workspace *workspace)
{
    struct view *view;
    struct view_proxy *view_proxy;
    /* find topmost enabled(mapped and not minimized) view and activate it */
    wl_list_for_each(view_proxy, &workspace->view_proxies, workspace_link) {
        view = view_proxy->view;
        if (!view->base.activatable || !view->base.mapped || view->base.minimized) {
            continue;
        }
        view_activate(view);
        seat_focus_surface(input_manager_get_default_seat(), view->surface);
        if (view->base.fullscreen) {
            view_reparent_fullscreen(view, true);
        }
        return;
    }

    /* workaround to hide fullscreen view when no view in workspace */
    view = view_manager->activated.view;
    if (view && view->base.fullscreen) {
        view_proxy = view_proxy_by_workspace(view, workspace);
        if (!view_proxy) {
            view_reparent_fullscreen(view, false);
            view_set_activated(view, false);
        }
    }

    /* no view can be activated, clear keyboard focus */
    seat_focus_surface(input_manager_get_default_seat(), NULL);
}

static void view_activate_with_workspace(struct view *view, bool find_parent)
{
    if (!view) {
        return;
    }
    if (view->parent && find_parent) {
        view_activate_with_workspace(view->parent, true);
    }
    struct view_proxy *view_proxy;
    wl_list_for_each(view_proxy, &view->view_proxies, view_link) {
        wl_list_remove(&view_proxy->workspace_link);
        wl_list_insert(&view_proxy->workspace->view_proxies, &view_proxy->workspace_link);
        ky_scene_node_raise_to_top(ky_scene_node_from_tree(view_proxy->tree));
    }
    /* raise children if any */
    struct view *child;
    wl_list_for_each(child, &view->children, parent_link) {
        view_activate_with_workspace(child, false);
    }
}

static void view_activate_without_workspace(struct view *view, bool find_parent)
{
    if (!view) {
        return;
    }
    if (view->parent && find_parent) {
        view_activate_without_workspace(view->parent, true);
    }
    ky_scene_node_raise_to_top(ky_scene_node_from_tree(view->tree));
    /* raise children if any */
    struct view *child;
    wl_list_for_each(child, &view->children, parent_link) {
        view_activate_without_workspace(child, false);
    }
}

void kywc_view_activate(struct kywc_view *kywc_view)
{
    struct view *view = view_from_kywc_view(kywc_view);
    view_activate(view);

    /* insert view proxy in workspace topmost */
    if (view->current_proxy) {
        view_activate_with_workspace(view, true);
    } else {
        view_activate_without_workspace(view, true);
    }
}

void kywc_view_set_tiled(struct kywc_view *kywc_view, enum kywc_tile tile,
                         struct kywc_output *kywc_output)
{
    if (kywc_view->fullscreen || !kywc_view->resizable) {
        return;
    }

    struct view *view = view_from_kywc_view(kywc_view);

    /* tiled mode may switch between outputs */
    if (kywc_view->tiled == tile && (!kywc_output || kywc_output == view->output)) {
        return;
    }

    struct kywc_box geo = { 0 };
    view_get_tiled_geometry(view, &geo, kywc_output ? kywc_output : view->output, tile);

    /* may switch between tiled modes */
    if (kywc_view->tiled == KYWC_TILE_NONE && tile != KYWC_TILE_NONE) {
        if (!kywc_view->maximized) {
            view->saved.geometry = view->base.geometry;
        }
    }

    if (kywc_view->maximized) {
        view->pending.action |= VIEW_ACTION_MAXIMIZE;
        kywc_view->maximized = false;
    }

    kywc_view->tiled = tile;
    view->pending.action |= VIEW_ACTION_TILE;
    view->pending.geometry = geo;

    if (kywc_view->mapped && view->impl->configure) {
        view->impl->configure(view);
    }
}

void kywc_view_set_minimized(struct kywc_view *kywc_view, bool minimized)
{
    if (!kywc_view->minimizable || kywc_view->minimized == minimized) {
        return;
    }

    struct view *view = view_from_kywc_view(kywc_view);
    ky_scene_node_set_enabled(ky_scene_node_from_tree(view->tree), !minimized);

    kywc_view->minimized = minimized;
    view->pending.action |= VIEW_ACTION_MINIMIZE;

    if (kywc_view->mapped && view->impl->configure) {
        view->impl->configure(view);
    }

    /* if view is the activated view, process it in activated.minimize listener */
    input_rebase_all_cursor();
    wl_signal_emit_mutable(&kywc_view->events.minimize, NULL);

    if (!kywc_view->minimized && view_manager->show_desktop_enabled) {
        view_manager_show_desktop(false, false);
    }
}

void kywc_view_toggle_minimized(struct kywc_view *kywc_view)
{
    kywc_view_set_minimized(kywc_view, !kywc_view->minimized);
}

void kywc_view_set_maximized(struct kywc_view *kywc_view, bool maximized,
                             struct kywc_output *kywc_output)
{
    struct view *view = view_from_kywc_view(kywc_view);

    /* tiled to unmaximized after tiled from maximized */
    if (!kywc_view->maximizable || (!kywc_view->tiled && kywc_view->maximized == maximized &&
                                    (!kywc_output || kywc_output == view->output))) {
        return;
    }

    struct kywc_box geo = { 0 };

    if (maximized) {
        view_get_tiled_geometry(view, &geo, kywc_output ? kywc_output : view->output,
                                KYWC_TILE_ALL);
        if (!kywc_view->tiled) {
            view->saved.geometry = kywc_view->geometry;
        }
    } else {
        geo = view->saved.geometry;
    }

    /* don't restore tiled mode followed other compositors */
    if (kywc_view->tiled) {
        view->pending.action |= VIEW_ACTION_TILE;
        kywc_view->tiled = KYWC_TILE_NONE;
    }

    kywc_view->maximized = maximized;
    view->pending.action |= VIEW_ACTION_MAXIMIZE;
    view->pending.geometry = geo;

    if (kywc_view->mapped && view->impl->configure) {
        view->impl->configure(view);
    }
}

void kywc_view_toggle_maximized(struct kywc_view *kywc_view)
{
    kywc_view_set_maximized(kywc_view, !kywc_view->maximized, NULL);
}

static void view_reparent_fullscreen(struct view *view, bool active)
{
    struct kywc_view *kywc_view = &view->base;
    struct view_proxy *view_proxy = NULL;
    struct view_layer *layer = NULL;
    if (active) {
        view->saved.layer = kywc_view->kept_above
                                ? LAYER_ABOVE
                                : (kywc_view->kept_below ? LAYER_BELOW : LAYER_NORMAL);
        layer = view_manager_get_layer(LAYER_ACTIVE, false);
    }

    wl_list_for_each(view_proxy, &view->view_proxies, view_link) {
        if (!active) {
            /* restore fullscreen view to workspace */
            layer = workspace_layer(view_proxy->workspace, view->saved.layer);
        }
        ky_scene_node_reparent(ky_scene_node_from_tree(view_proxy->tree), layer->tree);
    }
}

void kywc_view_set_fullscreen(struct kywc_view *kywc_view, bool fullscreen,
                              struct kywc_output *kywc_output)
{
    struct view *view = view_from_kywc_view(kywc_view);

    if (!kywc_view->fullscreenable ||
        (kywc_view->fullscreen == fullscreen && (!kywc_output || kywc_output == view->output))) {
        return;
    }

    struct kywc_box geo = { 0 };

    if (fullscreen) {
        kywc_output_effective_geometry(kywc_output ? kywc_output : view->output, &geo);
        if (!kywc_view->maximized && !kywc_view->tiled) {
            view->saved.geometry = kywc_view->geometry;
        }
    } else if (kywc_view->maximized) {
        view_get_tiled_geometry(view, &geo, view->output, KYWC_TILE_ALL);
    } else if (kywc_view->tiled) {
        view_get_tiled_geometry(view, &geo, view->output, kywc_view->tiled);
    } else {
        geo = view->saved.geometry;
    }

    view_reparent_fullscreen(view, fullscreen);
    kywc_view->fullscreen = fullscreen;

    view->pending.action |= VIEW_ACTION_FULLSCREEN;
    view->pending.geometry = geo;

    if (kywc_view->mapped && view->impl->configure) {
        view->impl->configure(view);
    }
}

void kywc_view_toggle_fullscreen(struct kywc_view *kywc_view)
{
    kywc_view_set_fullscreen(kywc_view, !kywc_view->fullscreen, NULL);
}

void kywc_view_set_kept_above(struct kywc_view *kywc_view, bool kept_above)
{
    struct view *view = view_from_kywc_view(kywc_view);
    assert(view->current_proxy);

    if (kywc_view->kept_above == kept_above) {
        return;
    }

    kywc_view->kept_above = kept_above;
    kywc_view->kept_below = false;

    enum layer layer = kept_above ? LAYER_ABOVE : LAYER_NORMAL;
    struct view_proxy *proxy;
    wl_list_for_each(proxy, &view->view_proxies, view_link) {
        struct view_layer *view_layer = workspace_layer(proxy->workspace, layer);
        ky_scene_node_reparent(ky_scene_node_from_tree(proxy->tree), view_layer->tree);
    }
}

void kywc_view_toggle_kept_above(struct kywc_view *kywc_view)
{
    kywc_view_set_kept_above(kywc_view, !kywc_view->kept_above);
}

void kywc_view_set_kept_below(struct kywc_view *kywc_view, bool kept_below)
{
    struct view *view = view_from_kywc_view(kywc_view);
    assert(view->current_proxy);

    if (kywc_view->kept_below == kept_below) {
        return;
    }

    kywc_view->kept_below = kept_below;
    kywc_view->kept_above = false;

    enum layer layer = kept_below ? LAYER_BELOW : LAYER_NORMAL;
    struct view_proxy *proxy;
    wl_list_for_each(proxy, &view->view_proxies, view_link) {
        struct view_layer *view_layer = workspace_layer(proxy->workspace, layer);
        ky_scene_node_reparent(ky_scene_node_from_tree(proxy->tree), view_layer->tree);
    }
}

void kywc_view_toggle_kept_below(struct kywc_view *kywc_view)
{
    kywc_view_set_kept_below(kywc_view, !kywc_view->kept_below);
}

void view_helper_move(struct view *view, int x, int y)
{
    struct kywc_box *geo = &view->base.geometry;
    bool changed = geo->x != x || geo->y != y;

    geo->x = x;
    geo->y = y;

    /* if we have two monitor and place a window on the left one,
     * then unplug the left one, window will be placed to the right one by positioner.
     * the postion will not change as the right one become (0,0).
     */
    view_update_output(view);

    if (changed) {
        ky_scene_node_set_position(ky_scene_node_from_tree(view->tree), x, y);
        wl_signal_emit_mutable(&view->base.events.position, NULL);
    }
}

bool view_is_moveable(struct view *view)
{
    struct kywc_view *kywc_view = &view->base;
    return kywc_view->movable && !kywc_view->fullscreen;
}

bool view_is_resizable(struct view *view)
{
    struct kywc_view *kywc_view = &view->base;
    return kywc_view->resizable && !kywc_view->fullscreen && !kywc_view->maximized;
}

void view_update_size(struct view *view, int width, int height, int min_width, int min_height,
                      int max_width, int max_height)
{
    struct kywc_view *kywc_view = &view->base;

    if (kywc_view->min_width != min_width || kywc_view->min_height != min_height) {
        kywc_view->min_width = min_width;
        kywc_view->min_height = min_height;
        kywc_log(KYWC_DEBUG, "view %p minimal size to %d x %d", view, min_width, min_height);
    }

    if (kywc_view->max_width != max_width || kywc_view->max_height != max_height) {
        kywc_view->max_width = max_width;
        kywc_view->max_height = max_height;
        kywc_log(KYWC_DEBUG, "view %p maximal size to %d x %d", view, max_width, max_height);
    }

    if (kywc_view->max_width != 0 && kywc_view->max_height != 0 &&
        kywc_view->min_width == kywc_view->max_width &&
        kywc_view->min_height == kywc_view->max_height) {
        kywc_view->maximizable = false;
        kywc_view->resizable = false;
    }

    if (kywc_view->geometry.width != width || kywc_view->geometry.height != height) {
        kywc_view->geometry.width = width;
        kywc_view->geometry.height = height;

        view_update_output(view);

        kywc_log(KYWC_DEBUG, "view %p size to %d x %d", view, width, height);
        wl_signal_emit_mutable(&view->base.events.size, NULL);
    }
}

void view_close_popups(struct view *view)
{
    if (view->impl->close_popups) {
        view->impl->close_popups(view);
    }
}

void view_show_window_menu(struct view *view, struct seat *seat, int x, int y)
{
    /* still show window menu even if no ssd used */
    if (!view->current_proxy) {
        return;
    }
    struct view_show_window_menu_event event = { view, seat, x, y };
    wl_signal_emit_mutable(&view_manager->events.window_menu, &event);
}

void view_manager_show_desktop(bool enabled, bool apply)
{
    if (view_manager->show_desktop_enabled == enabled) {
        return;
    }
    view_manager->show_desktop_enabled = enabled;

    /* minimize all view in current workspace */
    struct workspace *workspace = workspace_manager_get_current();
    struct view_proxy *view_proxy;
    wl_list_for_each_reverse(view_proxy, &workspace->view_proxies, workspace_link) {
        struct view *view = view_proxy->view;
        /* skip views not mapped */
        if (!view->base.mapped) {
            continue;
        }
        /* skip restoring views not minimized by show desktop */
        if (!enabled && !view->minimized_when_show_desktop) {
            continue;
        }
        /* true only the view is not minimized when going show desktop */
        view->minimized_when_show_desktop = enabled && !view->base.minimized;
        /* don't restoring views if the state is breaked */
        if (apply) {
            kywc_view_set_minimized(&view->base, enabled);
        }
    }

    if (apply && !enabled) {
        view_topmost_activate(workspace);
    }

    wl_signal_emit_mutable(&view_manager->events.show_desktop, NULL);
}

void view_manager_add_show_desktop_listener(struct wl_listener *listener)
{
    wl_signal_add(&view_manager->events.show_desktop, listener);
}

bool view_manager_get_show_desktop(void)
{
    return view_manager->show_desktop_enabled;
}

static void handle_server_destroy(struct wl_listener *listener, void *data)
{
    wl_list_remove(&view_manager->server_destroy.link);

    free(view_manager);
    view_manager = NULL;
}

struct view_manager *view_manager_create(struct server *server)
{
    view_manager = calloc(1, sizeof(struct view_manager));
    if (!view_manager) {
        return NULL;
    }

    view_manager->server = server;
    wl_signal_init(&view_manager->events.new_view);
    wl_signal_init(&view_manager->events.window_menu);
    wl_signal_init(&view_manager->events.show_desktop);

    view_manager->server_destroy.notify = handle_server_destroy;
    server_add_destroy_listener(server, &view_manager->server_destroy);

    view_manager->activated.minimize.notify = handle_activated_view_minimized;
    view_manager->activated.unmap.notify = handle_activated_view_unmap;

    /* create all layers */
    for (int layer = LAYER_FIRST; layer < LAYER_NUMBER; layer++) {
        view_manager->layers[layer].layer = layer;
        view_manager->layers[layer].tree = ky_scene_create_tree(server->scene);
    }

    workspace_manager_create(view_manager);
    decoration_manager_create(view_manager);
    server_decoration_manager_create(view_manager);
    positioner_manager_create(view_manager);
    shadow_manager_create(view_manager);
    window_actions_create(view_manager);
    window_menu_manager_create(view_manager);
    view_manager_actions_create(view_manager);
    maximize_switcher_create(view_manager);

    xdg_shell_init(view_manager);

    wlr_layer_shell_manager_create(server);
    wlr_foreign_toplevel_manager_create(server);
    kde_plasma_shell_create(server);
    kde_plasma_window_management_create(server);
    kde_blur_manager_create(server);

    return view_manager;
}
